/*
 * drive.c - Simple abstractions for working with file systems and drives
 * Copyright (C) 2024  Alexander Rosenberg
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version. See the LICENSE file for more information.
 */
#include "drive.h"
#include "util.h"

#include <ctype.h>
#include <err.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <string.h>
#include <dirent.h>
#include <sys/param.h>
#include <sys/mount.h>
#include <pthread.h>

static int compare_drives(const Drive *d1, const Drive *d2) {
    return strcmp(d1->name, d2->name);
}

Drive *drive_list_usb_drives(size_t *count) {
    DIR *dir = opendir("/dev");
    if (!dir) {
        warn("failed to open /dev");
        return NULL;
    }
    Drive *drives = NULL;
    *count = 0;
    struct dirent *sd;
    while ((sd = readdir(dir))) {
        if (strncmp(sd->d_name, "da", 2) == 0 && isdigit(sd->d_name[2])) {
            drives = realloc_checked(drives, sizeof(Drive) * ++(*count));
            Drive *cur = &drives[*count - 1];
            cur->name = strdup_checked(sd->d_name);
            asprintf_checked(&cur->path, "/dev/%s", sd->d_name);
            cur->fs = drive_guess_fs(cur->path);
            cur->mnt = NULL;
        }
    }
    closedir(dir);
    if (*count) {
        qsort(drives, *count, sizeof(Drive),
              (int(*)(const void *, const void *))compare_drives);
    }
    return drives;
}

void drive_free_drives(Drive *drives, size_t count) {
    for (size_t i = 0; i < count; ++i) {
        free(drives[i].name);
        free(drives[i].path);
        free(drives[i].fs);
        free(drives[i].mnt);
    }
    free(drives);
}

static void close_pipe(void *ptr) {
    int *pipe = ptr;
    if (pipe[0] >= 0) {
        close(pipe[0]);
    }
    if (pipe[1] >= 0) {
        close(pipe[1]);
    }
}

char *drive_guess_fs(const char *dev) {
    char *result_fs = NULL;
    int ctop[2] = {0, 0};
    pthread_cleanup_push(close_pipe, ctop);
    pthread_cleanup_push(free, result_fs);
    if (pipe(ctop) < 0) {
        warn("failed to create pipe");
        goto cleanup;
    }
    pid_t pid = fork();
    if (pid < 0) {
        warn("fork(2) failed");
        goto cleanup;
    } else if (pid == 0) {
        // child
        close(ctop[0]); // close read end
        if (dup2(ctop[1], 1) < 0) {
            warn("dup2(2) failed");
            close(ctop[1]);
            exit(1);
        }
        if (!GLOBAL_OPTS.verbose) {
            freopen("/dev/null", "w", stderr); // prevent warnings for unknown fs
        }
        execl("/usr/sbin/fstyp", "/usr/sbin/fstyp", dev, NULL);
        close(ctop[1]);
        exit(1);
    } 
    // parent
    int status;
    if (waitpid(pid, &status, 0) < 0) {
        warn("waiting for child failed");
        goto cleanup;
    }
    if (WEXITSTATUS(status) != 0) {
        // no error, we probably just don't know the FS type
        goto cleanup;
    }
    char read_buf[32];
    ssize_t read_count = read(ctop[0], read_buf, 32);
    if (read_count < 0) {
        warn("reading from pipe failed");
        goto cleanup;
    }
    read_buf[read_count - 1] = '\0';
    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
    result_fs = strdup_checked(read_buf);
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
 cleanup:
    pthread_cleanup_pop(false); // dont free result
    pthread_cleanup_pop(true);
    return result_fs;
}

bool drive_newfs_msdos(Drive *drive) {
    pid_t pid = fork();
    if (pid < 0) {
        warn("fork(2) failed");
        return false;
    } else if (pid == 0) {
        // child
        if (!GLOBAL_OPTS.verbose) {
            freopen("/dev/null", "w", stdout);
        }
        execl("/sbin/newfs_msdos", "/sbin/newfs_msdos", "-F32", drive->path, NULL);
        warn("execl(2) failed");
        exit(1);
    } else {
        // parent
        int status;
        if (waitpid(pid, &status, 0) < 0) {
            warn("waiting for child failed");
            return false;
        }
        if (WEXITSTATUS(status) != 0) {
            warnx("creating FAT32 FS failed");
            return false;
        }
        pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
        free(drive->fs);
        drive->fs = strdup_checked("msdosfs");
        pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
        return true;
    }
}

bool drive_mount(Drive *drive) {
    bool status = true;
    char *name = NULL;
    pthread_cleanup_push(free, name);
    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
    name = strdup_checked("/tmp/mntXXXXXX");
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
    if (!mkdtemp(name)) {
        warn("failed to create temp mount point");
        status = false;
        goto cleanup;
    }
    pid_t pid = fork();
    if (pid < 0) {
        warn("fork(2) failed");
        status = false;
        goto cleanup;
    } else if (pid == 0) {
        // child
        if (!GLOBAL_OPTS.verbose) {
            freopen("/dev/null", "w", stdout);
        }
        execl("/sbin/mount", "/sbin/mount", "-t", drive->fs, drive->path, name, NULL);
        warn("execl(2) failed");
        abort();
    }
    // parent
    int exit_code;
    if (waitpid(pid, &exit_code, 0) < 0) {
        warn("waiting for child failed");
        rmdir(name);
        status = false;
        goto  cleanup;
    }
    if (WEXITSTATUS(exit_code) != 0) {
        warnx("mount(8) failed: \"%s\"", drive->path);
        rmdir(name);
        status = false;
        goto  cleanup;
    }
    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
    drive->mnt = strdup_checked(name);
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
    LOG_VERBOSE("Mounted \"%s\" with fs %s to \"%s\"\n", drive->path, drive->fs,
                drive->mnt);
 cleanup:
    pthread_cleanup_pop(true);
    return status;
}

bool drive_unmount(Drive *drive, bool force) {
    if (!drive->mnt) {
        warnx("\"%s\" is not mounted", drive->path);
        return false;
    }
    int flags = force ? MNT_FORCE : 0;
    if (unmount(drive->mnt, flags) < 0) {
        warn("unmount(2) failed for \"%s\"", drive->path);
        return false;
    }
    rmdir(drive->mnt);
    pthread_setcancelstate(PTHREAD_CANCEL_DISABLE, NULL);
    free(drive->mnt);
    drive->mnt = NULL;
    pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
    LOG_VERBOSE("Unmounted \"%s\"\n", drive->path);
    return true;
}
